# coding=utf8

from context import get_json_value
import base64
import requests
from util import AssetStatus, AssetInfoFailedException, print_traceback
from event import log_event, AUTO_KEYWORDER_ERROR


def get_asset_info(obj, objecttype, asset_field, asset_version, url_required):
    assets = get_json_value(obj, '%s.%s' % (objecttype, asset_field))
    if not isinstance(assets, list) or len(assets) < 1:
        raise AssetInfoFailedException(
            'asset field %s.%s' % (objecttype, asset_field))

    eas_id = get_json_value(assets[0], '_id')
    if eas_id is None:
        raise AssetInfoFailedException(
            'versions.%s._id not found in asset field' % asset_version)

    eas_url = None
    if url_required:
        eas_url = get_json_value(assets[0],
                                 'versions.%s.url' % asset_version)
        if eas_url is None:
            raise AssetInfoFailedException(
                'versions.%s.url not found in asset field' % asset_version)

    img_type = get_json_value(assets[0],
                              'versions.%s.extension' % asset_version)
    if img_type is None:
        raise AssetInfoFailedException(
            'versions.%s.extension not found in asset field' % asset_version)

    original_filepath = get_json_value(assets[0], 'original_filepath')

    return eas_id, eas_url, img_type, original_filepath


def map_assets(easydb_context, asset_map, objecttype, asset_field, asset_version):
    logger = easydb_context.get_logger('pf.plugin.base.auto_keyworder.eas')

    pending_objects = set()

    for sys_id in asset_map:
        asset = asset_map[sys_id]
        if asset.eas_id is not None:
            continue
        if asset.easydb_obj is None:
            continue

        try:
            eas_id, eas_url, img_type, original_filepath = get_asset_info(asset.easydb_obj,
                                                                          objecttype,
                                                                          asset_field,
                                                                          asset_version,
                                                                          True)
            logger.debug('obj %s #%d: asset: path: %s, id: %d, version: %s, url: %s, type: %s' %
                         (objecttype, sys_id, original_filepath, eas_id, asset_version, eas_url, img_type))

            base64_data = base64_from_url(eas_url)
            logger.debug('obj %s #%d: asset: path: %s, id: %d, base64 data: %d bytes' %
                         (objecttype, sys_id, original_filepath, eas_id, len(base64_data)))
            if base64_data is None:
                continue

            asset_map[sys_id].eas_id = eas_id
            asset_map[sys_id].base64_data = base64_data
            asset_map[sys_id].img_type = img_type
            asset_map[sys_id].original_filepath = original_filepath

            pending_objects.add(sys_id)
        except AssetInfoFailedException as e:
            logger.debug('could not get asset information from object %s #%d: %s' % (
                objecttype, sys_id, e.detail
            ))
            log_event(easydb_context,
                      AUTO_KEYWORDER_ERROR,
                      {
                          'error': 'could not get asset information',
                          'details': e.detail,
                          'asset': {
                              'eas_id': asset_map[sys_id].eas_id,
                              'extension': asset_map[sys_id].img_type,
                              'original_filepath': asset_map[sys_id].original_filepath,
                              'version': asset_version
                          },
                          'object': {
                              '_objecttype': objecttype,
                              '_id': get_json_value(asset_map[sys_id].easydb_obj, '%s._id' % objecttype),
                              '_system_object_id': sys_id
                          }
                      })

        except Exception as e:
            logger.debug('could not get asset information from object %s #%d: %s' % (
                objecttype, sys_id, e.message
            ))
            print_traceback(logger, e)

    return pending_objects


def base64_from_url(fileurl):
    response = requests.get(fileurl)
    if response.status_code != 200:
        return None

    return base64.b64encode(response.content).decode('utf-8')
